#include "lcdgfx.h"
#include "scroll_phat_hd.h"
#include "lcd_font.h"

#include <string.h>
#include <stdio.h>

static uint8_t *_lcd_fb = 0;
static int _lcd_w = 0;
static int _lcd_h = 0;

void lcd_init() {
	phat_hd_open();

	_lcd_fb = 0;
	_lcd_w = 0;
	_lcd_h = 0;
}

void lcd_deinit() {
	if(_lcd_fb)
		delete [] _lcd_fb;
	_lcd_fb = 0;
	_lcd_w = 0;
	_lcd_h = 0;
	phat_hd_close();
}

int lcd_get_fb_width() {
	return _lcd_w;
}

int lcd_get_fb_height() {
	return _lcd_h;
}
#define PHAT_W 17
#define PHAT_H 7

#define min(x,y) ((x<y)?x:y)
#define max(x,y) ((x>y)?x:y)

static uint8_t _db[PHAT_W*PHAT_H]; 
static int _offset_x = 0;
static int _offset_y = 0;
static bool _wrap_x = false;
static bool _wrap_y = false;

void lcd_show(int buffer) {
	memset(_db, 0, sizeof(_db));
	uint8_t *p=_db;

#if 0
	printf("lcd w=%d h=%d\n", _lcd_w, _lcd_h);
	{
		uint8_t *p = _lcd_fb;
		for(int y=0;y<_lcd_h;y++) {
			for(int x=0;x<_lcd_w;x++) {
				printf("%3d ", (int)*p++);
			}
			printf("\n");
		}
		printf("\n");
	}
#endif
	for(int y=0;y<PHAT_H; y++) {
		for(int x=0;x<PHAT_W;x++) {
			int tx = x + _offset_x;
			int ty = y + _offset_y;

			if(_wrap_x) {
				tx = tx % _lcd_w;
			}
			if(_wrap_y) {
				ty = ty % _lcd_h;
			}
			*p = lcd_get_pixel(tx, ty);
//			printf("%3d ", (int)*p);
			p++;
		}
//		printf("\n");
	}

	phat_hd_draw(_db, buffer);
}

void lcd_animate(int delaySteps) {
	phat_hd_animation_mode(delaySteps);
}

void lcd_picture_mode() {
	phat_hd_picture_mode();
}

void lcd_scroll_to(int x, int y) {
	_offset_x = x;
	_offset_y = y;
}

void lcd_scroll(int x, int y) {
	_offset_x += x;
	_offset_y += y;
	if(_wrap_x) {
		_offset_x = _offset_x % _lcd_w;
	}
	if(_wrap_y) {
		_offset_y = _offset_y % _lcd_h;
	}
}

void lcd_wrap(bool wrap_x, bool wrap_y) {
	_wrap_x = wrap_x;
	_wrap_y = wrap_y;
}

// Set a new size for the LCD, while 
// preserving existing data, if any.
void lcd_set_size(int w, int h) {
	uint8_t *prev_fb = _lcd_fb;

//	printf("Resizing fb to %d %d\n", w, h);

	// Allocate & clear the new fb
	_lcd_fb = new uint8_t[w*h];
	memset(_lcd_fb, 0, w*h);

	if(prev_fb) {
		int rows_to_copy = min(h, _lcd_h);
		int cols_to_copy = min(w, _lcd_w);
		uint8_t *src = prev_fb;
		uint8_t *dst = _lcd_fb;

		for(int y=0;y<rows_to_copy;y++) {
			memcpy(dst, src, cols_to_copy);
			dst += w;
			src += _lcd_w; // stride of previous fb
		}

		// free old fb
		delete []prev_fb;
	}
	_lcd_w = w;
	_lcd_h = h;
}

void _ensure_fb_size(int x, int y) {
	int nw=max(x+1,_lcd_w);
	int nh=max(y+1,_lcd_h);
	if((nw > _lcd_w) || (nh > _lcd_h)) {
		lcd_set_size(nw, nh);
	}
}

// Drawing functions

inline uint8_t *_lcd_ptr(int x, int y) {
	return _lcd_fb + (y*_lcd_w) + x;
}

void lcd_set_pixel(int x, int y, float brightness) {
	_ensure_fb_size(x, y);

	uint8_t *p = _lcd_ptr(x, y);
	*p = (brightness*255.0);
}

void lcd_fill(float brightness, int x, int y, int w, int h) {
	_ensure_fb_size(x+w-1, y+h-1);

	uint8_t val = brightness*255.0;
	for(int i=0; i<h; i++) {
		uint8_t *p = _lcd_ptr(x, y+i);
		memset(p, val, w);
	}
}

void lcd_clear() {
	lcd_fill(0, 0, 0, _lcd_w, _lcd_h);
}

void lcd_draw_char(int x, int y, uint8_t ch, float brightness, bool reverse_video=false) {
	Font &font = font3x5;
	FontChar &fc = font.charData[ch];

	_ensure_fb_size(x+fc.width-1, y+fc.height-1);

//	printf("x = %d y= %d char = %d fc height = %d width=%d\n", x, y, (int) ch, fc.height, fc.width);
	const uint8_t *fd = fc.data;
	for(int j=0;j<fc.height;j++) {
		uint8_t *p = _lcd_ptr(x, (y+j));
		for(int i=0;i<fc.width;i++) {
			uint8_t pv = *fd++;
			if(reverse_video)
				pv = ~pv;
			pv = pv * brightness;
			//printf("%03d ", (int)pv);
			*p++ = pv;
		}
		//printf("\n");
	}
}

void lcd_draw_string(const char *str, int x, int y, float brightness, bool horizontal, bool reverse_video) {
	Font &font = font3x5;
	while(*str) {
		uint8_t ch=*str++;
		FontChar &fc = font.charData[ch];
//		printf("Char at %d %d\n", x, y);
		lcd_draw_char(x, y, ch, brightness, reverse_video);
		if(horizontal) {
			// character determines h spacing
			x += (fc.width + 1); 
		} else {
			// font determines v spacing
			y += (font.height + 1);
		}
	}
}

uint8_t lcd_get_pixel(int x, int y) {
	if((x<0)|| (y<0) || (x>=_lcd_w) || (y>=_lcd_h)) {
		return 0;
	}

	return *(_lcd_ptr(x,y));
}
