#include "scroll_phat_hd.h"
#include <string.h>
#include <stdint.h>

#include <linux/i2c.h>
#include <linux/i2c-dev.h>
#include <sys/ioctl.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <stdio.h>

const uint8_t D_WIDTH = 17;
const uint8_t D_HEIGHT = 7;

const uint8_t MODE_REGISTER = 0x00;
const uint8_t FRAME_REGISTER = 0x01;
const uint8_t AUTOPLAY1_REGISTER = 0x02;
const uint8_t AUTOPLAY2_REGISTER = 0x03;
const uint8_t BLINK_REGISTER = 0x05;
const uint8_t AUDIOSYNC_REGISTER = 0x06;
const uint8_t BREATH1_REGISTER = 0x08;
const uint8_t BREATH2_REGISTER = 0x09;
const uint8_t SHUTDOWN_REGISTER = 0x0a;
const uint8_t GAIN_REGISTER = 0x0b;
const uint8_t ADC_REGISTER = 0x0c;

const uint8_t CONFIG_BANK = 0x0b;
const uint8_t BANK_ADDRESS = 0xfd;

const uint8_t PICTURE_MODE = 0x00;
const uint8_t AUTOPLAY_MODE = 0x08;
const uint8_t AUDIOPLAY_MODE = 0x18;

const uint8_t ENABLE_OFFSET = 0x00;
const uint8_t BLINK_OFFSET = 0x12;
const uint8_t COLOR_OFFSET = 0x24;

static uint8_t deviceAddress = 0x74;
static int fd = -1;
static char fb[144];
static uint8_t current_frame = 1;

#define SUCCESS 1
#define FAIL 0

void _bank(int bank);
void _frame(int frame);
int _pixel_offset_in_bank(int x, int y);
int _write_reg(uint8_t reg, uint8_t value);
int _write_block(uint8_t reg, void *pdata, int count);
int _write_block_chunk(uint8_t reg, void *pdata, int count);

void _reset() {
	_bank(CONFIG_BANK);
	_write_reg(SHUTDOWN_REGISTER, 0);
	usleep(10);
	_write_reg(SHUTDOWN_REGISTER, 1);
}

int phat_hd_open() {
	if(fd!=-1)
		return FAIL;

	fd = open("/dev/i2c-1", O_RDWR); // We know it's on bus 1...
	if(fd<0) {
		perror("Unable to open i2c control file\n");
		return FAIL;
	}

	// To begin, reset!
	_reset();

	// Empty the frame buffer
	phat_hd_clear();

	// Configure registers through the configuration bank
	_bank(CONFIG_BANK);
	_write_reg(MODE_REGISTER, PICTURE_MODE);
	_write_reg(AUDIOSYNC_REGISTER, 0);

	// The pHAT has 17*7 LEDs. 
	// Enable all LEDs of all frames. 
	// (XXX The IS31FL3731 claims to support 144 LEDs, but 
	// 17*8 = 136. Solve this mystery later).
	char enableBlock[17];
	memset(enableBlock, 0xff, sizeof(enableBlock));
	for(int i=0;i<8;i++) {
		_bank(i);
		_write_block(0, enableBlock, sizeof(enableBlock));
	}

	return SUCCESS;
}

int phat_hd_close() {
	if(fd==-1)
		return FAIL;

	return SUCCESS;
}

int phat_hd_clear() {
	memset(fb, 0x0, sizeof(fb));
	return phat_hd_draw(fb);
}

int phat_hd_draw(void *buffer, int page) {
	if(fd==-1)
		return FAIL;

	char *p = (char*) buffer;
	// remap buffer contents to frame buffer
	for(int y=0;y<D_HEIGHT;y++) {
		for(int x=0;x<D_WIDTH;x++) {
			fb[_pixel_offset_in_bank(x,y)] = *p++;
		}
	}

	if(page==-1) {
		// Decide & switch to the next frame
		int next_frame;
		next_frame = current_frame ^ 1;

		_bank(next_frame);

		// Copy framebuffer contents
		// XXX: ensure I2C can handle large writes! - else this won't work
		_write_block(COLOR_OFFSET, fb, sizeof(fb)); 

		// Show the next frame
		_frame(next_frame);

		current_frame = next_frame;
	} else {
		_bank(page);
		_write_block(COLOR_OFFSET, fb, sizeof(fb)); 
		// Uncomment below line to show the frame
		// _frame(page); 
	}

	return SUCCESS;
}

void phat_hd_animation_mode(int delaySteps) {
	_bank(CONFIG_BANK);
	_write_reg(AUTOPLAY1_REGISTER, 0);	
	_write_reg(AUTOPLAY2_REGISTER, delaySteps);	
	_write_reg(MODE_REGISTER, AUTOPLAY_MODE);
}

void phat_hd_picture_mode() {
	_bank(CONFIG_BANK);
	_write_reg(MODE_REGISTER, PICTURE_MODE);
}
int _pixel_offset_in_bank(int x, int y) {
	//y = D_HEIGHT-(y+1);
	x = D_WIDTH-(x+1);
	if(x>8) {
		x = x-8;
		y = 6 - (y+8);
	} else {
		x = 8-x;
	}
	return (x*16)+y;

	// Don't pull your hair. The scroll pHAT HD has a weird pixel layout
	// In case you're curious, the above function printed over [17,7] yields:
	//
	// 128  112   96   80   64   48   32   16    0   14   30   46   62   78   94  110  126 
	// 129  113   97   81   65   49   33   17    1   13   29   45   61   77   93  109  125 
	// 130  114   98   82   66   50   34   18    2   12   28   44   60   76   92  108  124 
	// 131  115   99   83   67   51   35   19    3   11   27   43   59   75   91  107  123 
	// 132  116  100   84   68   52   36   20    4   10   26   42   58   74   90  106  122 
	// 133  117  101   85   69   53   37   21    5    9   25   41   57   73   89  105  121 
	// 134  118  102   86   70   54   38   22    6    8   24   40   56   72   88  104  120 
}

void _bank(int bank) {
	_write_reg(BANK_ADDRESS, bank);	
}

void _frame(int frame) {
	_bank(CONFIG_BANK);
	_write_reg(FRAME_REGISTER, frame);	
}

uint8_t xferBuff[1024];

int _write_reg(uint8_t reg, uint8_t value) {
	return _write_block(reg, &value, 1);
}

int _write_block(uint8_t reg, void *pdata, int count) {
	struct i2c_msg messages[1];
	messages[0].addr = deviceAddress;
	messages[0].flags = 0;
	messages[0].buf = xferBuff;

	// FIll transfer buffer
	xferBuff[0]=reg; // first byte is register address
	memcpy(xferBuff+1, pdata, count); // remaining are data
	messages[0].len = count+1;

	struct i2c_rdwr_ioctl_data packets;
	packets.msgs = messages;
	packets.nmsgs = 1;
	if(ioctl(fd, I2C_RDWR, &packets) < 0) {
		perror("Unable to send data");
		return FAIL;
	}
#if 0
	printf("write_i2c_block_data 74 %x %d ", reg, count);
	printf("[");
	for(int i=0;i<count;i++) {
		printf("%d,", (int)(*((unsigned char*)pdata+i)));
	}
	printf("]\n");
#endif
	return SUCCESS;
}

int _write_block_chunk(uint8_t reg, void *pdata, int count) {
	int chunkSize = 32;
	while(count>0) {
		int nw = (count>=chunkSize)?chunkSize:count;
		if(_write_block(reg, pdata, nw)==FAIL) 
			return FAIL;
		reg = reg+nw;
		pdata = ((char*) pdata)+nw;
		count -= nw;
	}
	return SUCCESS;
}
